#
# shinGETsu - P2P anonymous BBS
#
# (c) FukTommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: Gateway.pm,v 1.6 2004/05/10 14:01:12 fuktommy Exp $
#

package Shingetsu::Gateway;
use Shingetsu::Config;
use Shingetsu::Message;
use Shingetsu::Signature;
use Shingetsu::Util;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(str_encode str_decode file_encode file_decode rec escape xdie
		args argsFromMulti touch readNote post newElement
		addRecord margeRecord search postDeleteMessage
);

#
# configuration
#
$datadir	= $Shingetsu::Config::datadir;
$message	= new Shingetsu::Message();

#
# tips
#
sub str_encode($) {		# from YukiWiki
	my ($s) = @_;
	my $encoded = '';
	foreach my $ch (split(//, $s)) {
		if ($ch =~ /[A-Za-z0-9_.\/#]/) {
			$encoded .= $ch;
		} else {
			$encoded .= '%' . sprintf("%02X", ord($ch));
		}
	}
	return $encoded;
}
sub str_decode($) {		# from YukiWiki
	my ($s) = @_;
	$s =~ tr/+/ /;
	$s =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C", hex($1))/eg;
	return escape($s);
}
sub file_encode($$) {		# from YukiWiki
	my ($type, $key) = @_;
	my $enkey = '';
	foreach my $ch (split(//, $key)) {
		$enkey .= sprintf("%02X", ord($ch));
	}
	return "${type}_${enkey}";
}
sub file_decode($) {
	my($name) = @_;
	@_ = split /_/, $name, 2;
	$name = $_[1];
	$name =~ s/[0-9A-F][0-9A-F]/pack("C", hex($&))/eg;
	return escape($name);
}
sub rec($) {
	return Shingetsu::Util::rec($_[0]);
}
sub xdie(;@) {
	print "Content-Type: text/plain;\r\n",
	      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
	      "\r\n",
	      @_, "\n";
	exit;
}

#
# escape
#
sub escape($) {
	local($_) = @_;
	s/&/&amp;/g;
	s/&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);/&$1;/g;
	s/</&lt;/g;
	s/>/&gt;/g;
	s/\r//g;
	s/\n/<br>/g;
	return $_;
}

#
# unescape arguments
#
sub args($) {
	local $_ = $_[0];
	local @_ = split /&/;
	my %arg = ();
	foreach (@_) {
		local @_ = split /=/;
		if (defined $_[1]) {
			$_[1] =~ s/\+/ /g;
			$_[1] =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
			$arg{$_[0]} = escape($_[1]);
		} else {
			$arg{$_[0]} = "";
		}
	}
	return %arg;
}

#
# unescape arguments(multipart/form-data)
#
sub argsFromMulti($) {
	my($input) = @_;
	local $_;
	my %arg = ();
	$input =~ /\s/;
	my $boundary = $`;
	my @input = split $boundary, $input;
	shift @input;
	foreach (@input) {
		my @buf = split /\r\n/, $_;
		next if (@buf == 1);
		shift @buf;
		$buf[0] =~ /Content-Disposition: form-data; name="([^"]+)"/i or next;
		my $key = $1;

		if ($buf[0] =~ /filename="([^"]+)"/i) {
			$arg{auto_suffix} = $1;
			$arg{auto_suffix} =~ s/.*[\/\\]//;
			if ($arg{auto_suffix} =~ /\.([^.]*)$/) {
				$arg{auto_suffix} = $1;
				$arg{auto_suffix} = lc $arg{auto_suffix};
			} else {
				$arg{auto_suffix} = "";
			}
		}

		$_ = shift @buf until ((! defined $_) || ($_ eq ""));
		$arg{$key} = join "\r\n", @buf;
		if ($key ne "attach") {
			$arg{$key} = escape($arg{$key});
		}
	}
	return %arg;
}

#
# make file
#
sub touch($) {
	my($file) = @_;
	my %stat = Shingetsu::CacheStat::list();
	unless (-s "$datadir/$file.dat") {
		Shingetsu::Util::lock();
		open OUT, ">> $datadir/$file.dat";
		close OUT;
		Shingetsu::Util::unlock();
		my %stat = Shingetsu::CacheStat::list();
		$stat{$file}{records} = 0;
		$stat{$file}{stamp}   = 0;
		$stat{$file}{size}    = 0;
		Shingetsu::CacheStat::sync(%stat);
	}
	if ((-s "$datadir/$file.dat") == 0) {
		my $node = Shingetsu::Cache::search($file);
		if (defined $node) {
			Shingetsu::Cache::getRegion($file, $node);
		}
	}
}

#
# read note history
#
sub readNote($) {
	my($file) = @_;
	local $_;
	my $buf = "";
	my %ref = ();
	if (open IN, "$datadir/$file.dat") {
		while (<IN>) {
			chomp;
			my %rec = rec($_);
			next if ((keys %rec <= 2) || (defined $rec{remove_stamp}));
			$buf = $_;
			$ref{"$rec{stamp}_$rec{id}"} = 0 if ($rec{base});
			if ($rec{base} && (defined $ref{$rec{base}})) {
				$ref{$rec{base}}++;
			}
		}
		close IN;
	}
	my $orphan = grep $ref{$_}==0, keys %ref;
	$orphan = 0 if ($orphan == 1);
	return ($buf, $orphan);
}

#
# add record
#
sub addRecord($$$) {
	my($file, $stamp, $body) = @_;
	my $id = Shingetsu::Extern::md5digest($body);
	Shingetsu::Util::lock();
	open OUT, ">> $datadir/$file.dat" or die "$datadir/$file.dat: failed to open. $!";
	print OUT "$stamp<>$id<>$body\n";
	close OUT;
	Shingetsu::Util::unlock();
	my %stat = Shingetsu::CacheStat::list();
	$stat{$file}{records}++;
	$stat{$file}{stamp} = $stamp;
	$stat{$file}{size}  = (-s "$datadir/$file.dat");
	Shingetsu::CacheStat::sync(%stat);
	return $id;
}

#
# post article
#
sub post($$) {
	my($thread, $input) = @_;
	local $_;
	my %arg = argsFromMulti($input);
	my ($suffix, $attach) = ("", "");
	$arg{name} = "" unless (defined $arg{name});
	$arg{mail} = "" unless (defined $arg{mail});
	$arg{passwd} = "" unless (defined $arg{passwd});
	if ((defined $arg{attach}) && ($arg{attach} ne "")) {
		if (length $arg{attach} > $Shingetsu::Config::filelimit*1024*1024) {
			xdie $$message{big_file};
		}
		$attach = Shingetsu::Extern::base64encode($arg{attach});
		if ($arg{suffix} ne "AUTO") {
			$suffix	= $arg{suffix};
		} elsif ((defined $arg{auto_suffix}) && ($arg{auto_suffix} ne "")) {
			$suffix	= $arg{auto_suffix};
		} else {
			$suffix	= "txt";
		}
	}
	my $stamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body  = "";
	my $target = "";
	$arg{suffix} = $suffix;
	$arg{attach} = $attach;
	foreach ("base", "name", "mail", "body", "suffix", "attach") {
		if ((defined $arg{$_}) && ($arg{$_} ne "")) {
			$body .= "<>$_:$arg{$_}";
			$target .= ",$_";
		}
	}
	$body =~ s/^<>//;
	$target =~ s/^,//;
	xdie $$message{null_article} if ($body eq "");
	if ($arg{passwd} ne "") {
		my($pubkey, $sign) = Shingetsu::Signature::generate($arg{passwd}, $body);
		$body = "pubkey:$pubkey<>sign:$sign<>target:$target<>" . $body;
	}
	my $id = addRecord($thread, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $thread);
		Shingetsu::Cache::tellupdate($thread, $stamp, $id, undef);
	}
	return substr $id, 0, 8;
}

#
# make new list element
#
sub newElement($$) {
	my($file, $input) = @_;
	local $_;
	my %arg = args($input);
	my $stamp = Shingetsu::Util::time();
	xdie $$message{null_name} if ($arg{name} eq "");
	xdie $$message{bad_name}  if ($arg{name} =~ m|[/#]|);
	xdie $$message{bad_name}  if ($arg{name} =~ m|\]\]|);
	xdie $$message{null_type} if ($arg{type} eq "");
	my $body = "type:$arg{type}<>link:$arg{name}";
	my $id = addRecord($file, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $file);
		Shingetsu::Cache::tellupdate($file, $stamp, $id, undef);
	}
}

#
# add new record and marge that and last record (for note)
#
sub margeRecord {
	my($file, $input) = @_;
	my %arg = args($input);
	my $stamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body = $arg{message};
	my $base = "";
	local $_;
	my($buf, $conflict) = readNote($file);
	if ($buf) {
		my %rec = rec($buf);
		if ($rec{body} ne "") {
			if ($rec{body} =~ /<br>$/) {
				$body = "$rec{body}<br>$body";
			} else {
				$body = "$rec{body}<br><br>$body";
			}
		}
		$base = "$rec{stamp}_$rec{id}";
	}
	xdie $$message{null_article} if ($body eq "");
	$body = "body:$body";
	$body .= "<>base:$base" if ($base);
	my $id = addRecord($file, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $file);
		Shingetsu::Cache::tellupdate($file, $stamp, $id, undef);
	}
	return substr $id, 0, 8;
}

#
# search string
# return: @filename
#
sub search($$) {
	my($type, $query) = @_;
	my $file;
	my @result = ();
	foreach $file (glob("$datadir/${type}_*.dat")) {
		my $file2 = $file;
		$file2 =~ s|^$datadir/||; $file2 =~ s|.dat$||;
		my $x = file_decode($file2);
		my $y = str_encode($x);
		my $found = 0;
		if ($x =~ /$query/) {
			$found = 1;
		} else {
			open IN, $file or die "$file: open failed. $!";
			while (<IN>) {
				if (/$query/) {
					$found = 1;
					last;
				}
			}
			close IN;
		}
		if ($found) {
			push @result, $file2;
		}
	}
	return @result;
}

#
# post delete message to other nodes
#
sub postDeleteMessage($$$%) {
	my($file, $stamp, $id, %arg) = @_;
	$arg{name} = "" unless (defined $arg{name});
	$arg{mail} = "" unless (defined $arg{mail});
	$arg{passwd} = "" unless (defined $arg{passwd});
	my $newstamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body  = "";
	my $target = "";
	if ((defined $arg{name})&&($arg{name} ne "")) {
		$body .= "<>name:$arg{name}";
		$target .= ",name";
	}
#	if ((defined $arg{mail})&&($arg{mail} ne "")) {
#		$body .= "<>mail:$arg{mail}";
#		$target .= ",mail";
#	}
	if ((defined $arg{message})&&($arg{message} ne "")) {
		$body .= "<>body:$arg{message}";
		$target .= ",body";
	}
	$body .= "<>remove_stamp:$stamp";	$target .= ",remove_stamp";
	$body .= "<>remove_id:$id";		$target .= ",remove_id";
	$body =~ s/^<>//;
	$target =~ s/^,//;
	xdie $$message{null_article} if ($body eq "");
	if ($arg{passwd} ne "") {
		my($pubkey, $sign) = Shingetsu::Signature::generate($arg{passwd}, $body);
		$body = "pubkey:$pubkey<>sign:$sign<>target:$target<>" . $body ;
	}
	my $newid = addRecord($file, $newstamp, $body);
	Shingetsu::Util::addUpdate($newstamp, $newid, $file);
	Shingetsu::Cache::tellupdate($file, $newstamp, $newid, undef);
}

1;
